//-----------------------------------------------------------------------------
/// Modulename : fwimage.cpp @file
///
/// Interface  : fwimage.h
///
///              Framework Image class. This is used by the framework to do the
///              most of the work for you. It handles the parameter passing
///              from the main application and includes some useful methods
///              to work with images.
///
///  2003 Sirona Dental Systems GmbH - All Rights Reserved
///
/// Status     : Draft
///
/// Security   : -
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Standard includes
//-----------------------------------------------------------------------------
#include "stdafx.h"

//-----------------------------------------------------------------------------
// Own includes
//-----------------------------------------------------------------------------
#include "resource.h"
#include "FwImage.h"
#include "isidexisng.h"

//-----------------------------------------------------------------------------
// Private definitions, -macros and -constants
//-----------------------------------------------------------------------------
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define LASTPARAMVERSION 1       ///< This is your own version of your parameter string.
                                 ///< You can use this to identify any changed parameter string in your future plugin versions

//-----------------------------------------------------------------------------
// Private data types
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// External references
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Module global variables
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Prototype of the private functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Implementation
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/// Constructor
///
/// Remarks  : -
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
CImage::CImage()
{
   m_pbmi      = NULL;
   m_hBitmap   = NULL;
}

//-----------------------------------------------------------------------------
/// Destructor
///
/// Remarks  : -
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
CImage::~CImage()
{
   Clear();
}

//-----------------------------------------------------------------------------
/// Clears internal buffers
///
/// Remarks  : -
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
void CImage::Clear()
{
   if (m_hBitmap)
   {
      (void) ::DeleteObject( m_hBitmap );
      m_hBitmap = NULL;
   }
   if (m_pbmi)
   {
      delete m_pbmi;
      m_pbmi = NULL;
   }
}

//-----------------------------------------------------------------------------
/// Pixel Set method. Used by framework
///
/// @param varImage [IN] The pixel data
/// @param nWidth [IN] Picture width
/// @param nHeight [IN] Picture height
///
/// @return TRUE if the image data has been successfully set
///
/// Remarks  : -
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
BOOL CImage::FromSidexis( LPDISPATCH pdispImage )
{
   if (!pdispImage)
      return FALSE;

   IImage         Image;
   Image.AttachDispatch( pdispImage, FALSE );

   //--------------------------------------------------------------------------
   // Clear internal buffers
   //--------------------------------------------------------------------------
   Clear();

   //--------------------------------------------------------------------------
   // Save the bitmap information internally
   //--------------------------------------------------------------------------
   m_pbmi = (BITMAPINFO*) new BYTE[ sizeof(BITMAPINFO) + 255 * sizeof( RGBQUAD ) ];   // maximum size

   m_pbmi->bmiHeader.biSize          = sizeof( BITMAPINFOHEADER );
   m_pbmi->bmiHeader.biWidth         = Image.GetWidth();
   m_pbmi->bmiHeader.biHeight        = Image.GetHeight();
   m_pbmi->bmiHeader.biPlanes        = 1;
   m_pbmi->bmiHeader.biBitCount      = Image.GetBitsPerPixel();
   m_pbmi->bmiHeader.biCompression   = BI_RGB;   // not compressed
   m_pbmi->bmiHeader.biSizeImage     = m_pbmi->bmiHeader.biWidth * m_pbmi->bmiHeader.biHeight * (m_pbmi->bmiHeader.biBitCount >> 3);
   m_pbmi->bmiHeader.biXPelsPerMeter = Image.GetXPelsPerMeter();
   m_pbmi->bmiHeader.biYPelsPerMeter = Image.GetYPelsPerMeter();
   m_pbmi->bmiHeader.biClrUsed       = 0;
   m_pbmi->bmiHeader.biClrImportant  = 0;

   //--------------------------------------------------------------------------
   // Palette data
   //--------------------------------------------------------------------------
   if (m_pbmi->bmiHeader.biBitCount == 8)
   {
      COleSafeArray sa;                     // Zum Auslesen des SafeArray
      long          lElements = 0;    // Zum Auslesen des SafeArray: Anzahl Bildpunkte
      long          lLBound   = 0;    // Zum Auslesen des SafeArray: Upper Bound
      long          lUBound   = 0;    // Zum Auslesen des SafeArray: Lower Bound
      COleVariant varPalette;

      (void) Image.GetPalette( &varPalette );

      RGBQUAD* pPaletteSrc = NULL; // Zum Auslesen des SafeArray: Palette

      sa.Attach( varPalette );

      sa.GetLBound( 1, &lLBound );
      sa.GetUBound( 1, &lUBound );
      sa.AccessData( (void**) &pPaletteSrc );// Get a pointer to the elements of the array and increments the lock count on the array

      lElements = lUBound - lLBound + 1;     // Get no. of elements in array

      // Inhalt kopieren
      (void) memcpy( m_pbmi->bmiColors, pPaletteSrc, sizeof( BYTE ) * lElements );

      sa.UnaccessData();                     // Decrement lock count
   }

   //--------------------------------------------------------------------------
   // ROI (has to be rotated for matching the actual rotation of image)
   //--------------------------------------------------------------------------
   LPDISPATCH pdispROI = Image.GetRoi();
   if (pdispROI)
   {
      IROI ROI;
      long lLeft = 0, lTop = 0, lWidth = 0, lHeight = 0;

      ROI.AttachDispatch( pdispROI );
      ROI.GetPosition( &lLeft, &lTop, &lWidth, &lHeight );
      m_rcROI.left   = lLeft;
      m_rcROI.top    = lTop;
      m_rcROI.right  = lLeft + lWidth - 1;
      m_rcROI.bottom = lTop + lHeight - 1;

      ROI.ReleaseDispatch();

      // The ROI coordinates have to be rotated to match the image rotation
      int   nRotation = (int) Image.GetRotation();
      CRect rcRotated; // Rckgabe

      switch (nRotation)
      {
         case  90:

            rcRotated.left   =               m_rcROI.top;
            rcRotated.top    = GetHeight() - m_rcROI.right;
            rcRotated.right  =               m_rcROI.bottom;
            rcRotated.bottom = GetHeight() - m_rcROI.left;

            break;

         case 180:

            rcRotated.left   = GetWidth()  - m_rcROI.right;
            rcRotated.top    = GetHeight() - m_rcROI.bottom;
            rcRotated.right  = GetWidth()  - m_rcROI.left;
            rcRotated.bottom = GetHeight() - m_rcROI.top;

            break;

         case 270:

            rcRotated.left   = GetWidth() - m_rcROI.bottom;
            rcRotated.top    =              m_rcROI.left;
            rcRotated.right  = GetWidth() - m_rcROI.top;
            rcRotated.bottom =              m_rcROI.right;

            break;

         default: 
            rcRotated = m_rcROI;
            break; // 0
      }

      m_rcROI = rcRotated;
   }
   else
   {
      m_rcROI.left   = 0;
      m_rcROI.top    = 0;
      m_rcROI.right  = GetWidth();
      m_rcROI.bottom = GetHeight();
   }

   //--------------------------------------------------------------------------
   // Generate the device-independent bitmap
   //--------------------------------------------------------------------------
   COleVariant    varPixels;
   COleSafeArray  saPixels;            // Passed in Pixel Array
   BYTE*          pPixels    = NULL;   // Pixel Array for copying
   BYTE*          pDIBPixels = NULL;   // The pixel data pointer returned from CreateDIBSection
   CDC            dc;                  // Needed as CreateDIBSection parameter

   varPixels.Clear();

   dc.CreateCompatibleDC( NULL );

   (void) Image.GetPixelMap( &varPixels );   // Get pixel data from SIDEXIS

   saPixels.Attach( varPixels );             // Access the SafeArray Bytes
   saPixels.AccessData((LPVOID*)&pPixels);

   // Creation of the HBITMAP to hold the image data
   m_hBitmap = CreateDIBSection( dc.GetSafeHdc(), m_pbmi, DIB_RGB_COLORS, (void**) &pDIBPixels, NULL, 0 );
   memcpy( pDIBPixels, pPixels, m_pbmi->bmiHeader.biSizeImage );

   saPixels.UnaccessData();   //decrement lock count

   if (!m_hBitmap)
      return FALSE;

   GdiFlush();

   //--------------------------------------------------------------------------
   // Detach so the caller can work with the dispatch further
   //--------------------------------------------------------------------------
   (void) Image.DetachDispatch();

   return TRUE;
}

//-----------------------------------------------------------------------------
/// Sets the pixel data to the given dispatch
///
/// @param     pdispImage [IN] Dispatch of image to set the pixels to
///
/// @return    TRUE on success
///
/// Remarks  : -
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
BOOL CImage::ToSidexis( LPDISPATCH pdispImage )
{
   IImage        Image;             // SIDEXIS Image COM object
   COleVariant   varPixelMap;       // Variant for passing data to Sidexis
   COleSafeArray sa;                // SafeArray
   long          lElements = 0;     // SafeArray parameter
   CDC           dc;                // Needed as parameter for GetDIBits
   BYTE        * pPixels = NULL;    // Needed for getting the bits out of HBITMAP

   VariantInit( &varPixelMap );

   if (!pdispImage)
      return FALSE;
   Image.AttachDispatch( pdispImage, FALSE );

   dc.CreateCompatibleDC( NULL );

   // Get the pixels back from the HBITMAP
   pPixels = new BYTE[ m_pbmi->bmiHeader.biSizeImage ];
   if (GetDIBits( dc.GetSafeHdc(), m_hBitmap, 0, GetHeight(), pPixels, m_pbmi, DIB_RGB_COLORS ) == 0)
      return FALSE;

   // Set the pixel map to SIDEXIS
   lElements = GetSize();

   sa.CreateOneDim( VT_UI1, lElements, pPixels );
   varPixelMap = sa.Detach();
   sa.Destroy();
   delete pPixels;
   pPixels = NULL;

   Image.SetPixelMap( varPixelMap );

   (void) Image.DetachDispatch();

   return TRUE;
}

//-----------------------------------------------------------------------------
/// Framework parameter passing
///
/// @param strParameters [IN] The macro parameters
///
/// @return TRUE if successful
///
/// Remarks  : Due to changes at the last parameter specifications a plug-in
///            has to return different last parameter strings. Versions prior 
///            to SIDEXIS XG 1.3 have to return the whole parameter string.
///            Newer versions only need the parameters needed by the plug-in
///            itself. The rest of the parameter string which the host needs
///            will be generated and saved by SIDEXIS XG.
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
BOOL CImage::SetParamsCompatible( const CString& strParameters )
{
   int      nPos = 0;
   CString  strTemp = strParameters;

   //--------------------------------------------------------------------------
   // Get the length of the parameters
   //--------------------------------------------------------------------------
   nPos = strTemp.Find( _T(",P") );

   if (nPos == -1)
      return FALSE;
   
   CString strOldParameter = strTemp.Right( strTemp.GetLength() - nPos - 2 );

   nPos = strOldParameter.Find( _T("=") );

   CString strLength = strOldParameter.Left( nPos );
   int     nLength   = _ttoi( strLength );

   strTemp = strOldParameter.Right( strOldParameter.GetLength() - nPos - 1 );

   //--------------------------------------------------------------------------
   // Cut off the parameters itself
   //--------------------------------------------------------------------------
   CString strFilterParameter = strTemp.Left( nLength );

   return SetParams( strFilterParameter );
}

//-----------------------------------------------------------------------------
/// Framework parameter passing
///
/// @return Returns the parameter String as needed by the main application
///
/// Remarks  : Due to changes at the last parameter specifications a plug-in
///            has to return different last parameter strings. Versions prior 
///            to SIDEXIS XG 1.3 have to return the whole parameter string.
///            Newer versions only need the parameters needed by the plug-in
///            itself. The rest of the parameter string which the host needs
///            will be generated and saved by SIDEXIS XG.
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
CString CImage::GetParamsCompatible()
{
   CString strParameters = GetParams();  // All parameters
   CString  strResult;                       // The return value

   //--------------------------------------------------------------------------
   // CLSID conversion
   //--------------------------------------------------------------------------
   CString  strCLSID;
   LPOLESTR pstr = NULL;

   StringFromCLSID( m_clsid, &pstr );

   strCLSID = pstr;

   //--------------------------------------------------------------------------
   // Format the whole string parameter as needed for SIDEXIS versions prior 1.3
   //--------------------------------------------------------------------------
   strResult.Format( _T("E1=1,I%d=%s,C1=3,V1=4,D1=5,P%d=%s"), 
                               strCLSID.GetLength(),      strCLSID,
                               strParameters.GetLength(), strParameters );

   return strResult;
}

//-----------------------------------------------------------------------------
/// Framework parameter passing
///
/// @param     strFilterParameters [IN] The filter parameters
///
/// @return    TRUE if successful
///
/// Remarks  : This method will be called by SIDEXIS XG to re-execute the
///            plug-in, e.g. for reconstructing an image view. The host
///            will set this parameter as returned by GetParams.
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
BOOL CImage::SetParams( const CString& strFilterParameters )
{
   //--------------------------------------------------------------------------
   // Rip off filter parameters and save them to our internal array
   //--------------------------------------------------------------------------
   CString strTemp = strFilterParameters;
   int     nPos    = strTemp.Find( _T(';') );

   //--------------------------------------------------------------------------
   // Get your own version identification
   //--------------------------------------------------------------------------
   CString strLastparamVersion = strTemp.Left( nPos );

   strTemp = strTemp.Right( strTemp.GetLength() - nPos - 1 );
   nPos = strTemp.Find( _T(';') );

   //--------------------------------------------------------------------------
   // Now rip off your filter parameters
   //--------------------------------------------------------------------------
   while (nPos != -1 || !strTemp.IsEmpty())
   {
      CString strValue;

      if (nPos != -1)
      {
         strValue = strTemp.Left( nPos );
         strTemp  = strTemp.Right( strTemp.GetLength() - nPos - 1 );
      }
      else
      {
         // Last Parameter
         strValue = strTemp;
         strTemp.Empty();
      }

      // Wert
      m_rgstrParameters.AddTail( strValue );

      nPos = strTemp.Find( _T(';') );
   }

   return TRUE;
}

//-----------------------------------------------------------------------------
/// Framework parameter passing
///
/// @return Returns the parameter String as needed by the main application
///
/// Remarks  : This method should return the parameter string which is needed
///            to re-execute the plugin function, e.g. for reconstructing an
///            image view. This string will be set unchanged through SetParams 
///            by SIDEXIS XG if the plug-in has to be executed for image view
///            reconstruction.
///
///            You should use ";" for separating the parameters
///
/// Security : -
///
/// Status   : Draft
///
/// Revision history - Version 1.0
///
/// 15.09.03 - Draft => V.1.0 (MCI, Fa. MGW)
//-----------------------------------------------------------------------------
CString CImage::GetParams()
{
   CString strParameters;                   // ; separated filter parameters

   //--------------------------------------------------------------------------
   // Save your last parameter version first for possible updates to this string from your side
   //--------------------------------------------------------------------------
   strParameters.Format( _T("%d;"), LASTPARAMVERSION );

   //--------------------------------------------------------------------------
   // Format the n Parameters to one String
   //--------------------------------------------------------------------------
   for (POSITION pos = m_rgstrParameters.GetHeadPosition(); pos; m_rgstrParameters.GetNext( pos ))
   {
      CString strParameterValue = m_rgstrParameters.GetAt( pos );
      CString strTemp;

      strTemp.Format( _T("%s;"), strParameterValue );

      strParameters += strTemp;
   }

   return strParameters;
}
